#!/usr/bin/perl

use strict;
use File::Tail;
use DBI;
use DBD::mysql;

my $db = "dnsrbl";
my $db_host = "127.0.0.1";
my $db_user = "dnsrbl";
my $db_pass = "password";
my $blackhole_domain = "dnsbl.example.com";
my $maillog = "/usr/local/psa/var/log/maillog";


my $log_file = "/var/spool/qscan/dnsbl.log";  #Where to log info - "" disables
my $block_score = 5;               #how many rejected emails before adding to dns block list (1 to 127)
my $clear_score = 10;              #how much extra slack emails known "good" senders can build to (0 to 128) - zero disables
my $process_interval = 86400;      #how long in seconds to wait intil we auto adjust scores - zero disables
my $decrement_ammt = 1;            #how much to reduce the "bad" senders score each interval - zero disables
my $increment_ammt = 1;            #how much to reduce the "good" senders bonus each interval - zero disables
my $remove_interval = 180;         #how long in days before deleting senders not seen from the DB - zero disables

# ********************************
# End of user configurable options
# ********************************

my $ip_number;
my $ip_tmp;
my $ip_addr;
my $rev_ip;
my $time_now;
my $time_epoch;
my $score;
my $logline;
my $interval;
my $blocked;
my $domain_name;
my @aged_scores;
my @oct;
my $i;

# Open logfile
if ($log_file){
  open LOGFILE , ">>$log_file" or die "Can't open log file: $!";
  select((select(LOGFILE), $| = 1)[0]);;
}

# Prep DB stuff
my $dsn="dbi:mysql:$db:$db_host:3306";
my $dbh = DBI->connect($dsn, $db_user, $db_pass) or die "Error: Can’t connect to the DB: $DBI::errstr\n";
my $get_score_sth  = $dbh->prepare('SELECT score, blocked FROM dnsrbl WHERE ip = ? LIMIT 1') or die "Error: Couldn't prepare statement: $DBI::errstr\n";
my $change_score_sth = $dbh->prepare('UPDATE dnsrbl SET score = ?, logline = ?, blocked = ? WHERE ip = ?') or die "Error: Couldn't prepare statement: $DBI::errstr\n";
my $new_entry_sth = $dbh->prepare('INSERT INTO dnsrbl (ip, score, added, processed, logline, blocked) VALUES (?, ?, ?, ?, ?, ?)') or die "Error: Couldn't prepare statement: $DBI::errstr\n";
my $aged_scores_sth = $dbh->prepare('SELECT ip, score, blocked FROM dnsrbl WHERE processed < ?') or die "Error: Couldn't prepare statement: $DBI::errstr\n";
my $processed_scores_sth = $dbh->prepare('UPDATE dnsrbl SET score = ?, processed = ?, blocked = ? WHERE ip = ?') or die "Error: Couldn't prepare statement: $DBI::errstr\n";
my $remove_senders_sth = $dbh->prepare('DELETE FROM dnsrbl WHERE processed < ?') or die "Error: Couldn't prepare statement: $DBI::errstr\n";

# Open maillog for watching
my $file=File::Tail->new("$maillog") or die "Error: Can’t tail the logfile: $!";
$clear_score = -($clear_score);

logit("Startup!\n");

while (defined($logline=$file->read)) {
  chomp($logline);
  if($logline =~ /qmail-scanner/) {

    #  Parse out IP
    $logline =~ /\b(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\b/;
    $ip_addr = "${1}.${2}.${3}.${4}";

    # Keep loopback out of reputation system
    if($ip_addr ne '127.0.0.1') {
      $ip_number=($1*16777216)+($2*65536)+($3*256)+$4;
      $rev_ip = "${4}.${3}.${2}.${1}";
      $time_epoch = time();
      $time_now= localtime();
      $get_score_sth->execute($ip_number) or die "Error: Couldn't execute statement: " . $get_score_sth->errstr;
      ($score, $blocked) = ($get_score_sth->fetchrow_array()) or $score=-9999;

      logit("${time_now}: Begin for IP ${ip_addr} the current score is ${score}\n");

      if($logline =~ /SPAM-REJECTED/ or $logline =~ /CLAMDSCAN/) {
        $score++;
        #sanity check
        if ($score > $block_score and $blocked == 0)  { $score = $block_score; }

        # Isn't in DB, so add it
        if ($score < -999) {
          $score = 1;
          $blocked = 0;
          $logline = $dbh->quote($logline);
          $new_entry_sth->execute($ip_number, $score, $time_epoch, $time_epoch, $logline, $blocked) or die "Error: Couldn't execute statement: " . $new_entry_sth->errstr;
        }

        # In DB and now ratcheted up to the blocking score and put in DNS
        elsif ($score == $block_score) {
          $blocked = 1;
          $logline = $dbh->quote($logline);
          $change_score_sth->execute($score, $logline, $blocked, $ip_number) or die "Error: Couldn't execute statement: " . $change_score_sth->errstr;

          # Add the entries to Plesk DNS
          if ($log_file) {
            logit("$time_now: ${rev_ip}.${blackhole_domain} is getting added to the block list.\n");
            system "/usr/local/psa/bin/dns -a ${blackhole_domain} -a ${rev_ip} -ip 127.0.0.2 >> $log_file";
          }
          else {
            system "/usr/local/psa/bin/dns -a ${blackhole_domain} -a ${rev_ip} -ip 127.0.0.2 >/dev/null 2>&1";
          }
        }

        # In DB, but not enough to block
        elsif ($score < $block_score) {
          $logline = $dbh->quote($logline);
          $change_score_sth->execute($score, $logline, $blocked, $ip_number) or die "Error: Couldn't execute statement: " . $change_score_sth->errstr;
        }
      }

      # Message is "good"
      if((($logline =~ /Clear/) and ($clear_score < 0)) and not ($logline =~ /\)\:SA\:1\(/)) {
        $logline = $dbh->quote($logline);

        # Isn't in DB, so add it
        if ($score == -9999) {
          $score = 0;
          $blocked = 0;
          $new_entry_sth->execute($ip_number, $score, $time_epoch, $time_epoch, $logline, $blocked) or die "Error: Couldn't execute statement: " . $new_entry_sth->errstr;;
        }
        # Wacky fun - under clear score, let's fix
        elsif ($score < $clear_score) {
          $score = $clear_score;
          $change_score_sth->execute($score, $logline, $blocked, $ip_number) or die "Error: Couldn't execute statement: " . $change_score_sth->errstr;
        }
        # decrement score for being a good sender
        elsif ($score > $clear_score) {
          $score--;
          $change_score_sth->execute($score, $logline, $blocked, $ip_number) or die "Error: Couldn't execute statement: " . $change_score_sth->errstr;
        }

        # I get a couple hundred messages per day that are legit.  For a more loaded
        # server, this would be better in a separate script run hourly. YMMV.

        # Do interval reputation processing
        $interval = $time_epoch - $process_interval;
        $aged_scores_sth->execute($interval) or die "Error: Couldn't execute statement: " . $aged_scores_sth->errstr;
        while(($ip_number, $score, $blocked)= $aged_scores_sth->fetchrow_array()) {
          if ($score != 0) {

            # DB Sanity Check
            if ($score > $block_score) { $score = $block_score; }  #$block_score is now lower - adjust db to reflect

            # OK, on to processing!
            elsif (($score > 0) and (($score-$decrement_ammt) <= 0)) {
              $score = 0;

              # Spent configured time in penalty box, so remove from dnsrbl and unblock
              if ($blocked == 1) {
                $blocked = 0;
                $ip_tmp = $ip_number;
                #IP Numer to quad  
                for($i = 3; $i >= 0; $i--) {
                  $oct[$i] = ($ip_tmp & 0xFF);
                  $ip_tmp >>= 8;
                }
                $rev_ip = "$oct[3].$oct[2].$oct[1].$oct[0]";
                logit("$time_now: ${rev_ip}.${blackhole_domain} is getting removed from block list.\n");
                if ($log_file) {
                  system "/usr/local/psa/bin/dns -d ${blackhole_domain} -a ${rev_ip} -ip 127.0.0.2 >> $log_file";
                }
                else {
                  system "/usr/local/psa/bin/dns -d ${blackhole_domain} -a ${rev_ip} -ip 127.0.0.2 >/dev/null 2>&1";
                }
              }
            }
            elsif ($score > 0) {
              $score = $score - $decrement_ammt;
            }
            elsif ($score < 0 and ($score + $increment_ammt) => 0) {
              $score = 0;
            }
            elsif ($score < 0) {
              $score = $score + $increment_ammt;
            }
            $processed_scores_sth->execute($score, $time_epoch, $blocked, $ip_number) or die "Error: Couldn't execute statement: " . $processed_scores_sth->errstr;
          }
        }
        # Remove entries that havent been seen again (stayed at 0 score)
        $interval = $time_epoch - ($remove_interval*86400);
        $remove_senders_sth->execute($interval) or die "Error: Couldn't execute statement: " . $remove_senders_sth->errstr;
      }
    }
  }
}

$dbh->disconnect();
if($log_file){close LOGFILE};

sub logit{
  if ($log_file){
        print LOGFILE "$_[0]";
  }
}
